from eth_abi import encode_single, decode_single
from eth_utils import function_signature_to_4byte_selector


def parse_signature(signature):
    """
    Breaks 'func(address)(uint256)' into ['func', '(address)', '(uint256)']
    """
    parts = []
    stack = []
    start = 0
    for end, letter in enumerate(signature):
        if letter == "(":
            stack.append(letter)
            if not parts:
                parts.append(signature[start:end])
                start = end
        if letter == ")":
            stack.pop()
            if not stack:  # we are only interested in outermost groups
                parts.append(signature[start : end + 1])
                start = end + 1
    return parts


class Signature:
    def __init__(self, signature):
        self.signature = signature
        self.parts = parse_signature(signature)
        self.input_types = self.parts[1]
        self.output_types = self.parts[2]
        self.function = "".join(self.parts[:2])
        self.fourbyte = function_signature_to_4byte_selector(self.function)

    def encode_data(self, args=None):
        return self.fourbyte + encode_single(self.input_types, args) if args else self.fourbyte

    def decode_data(self, output):
        return decode_single(self.output_types, output)



import typing as t

from web3.auto import w3
from lido.constants.contract_addresses import (
    get_lido_address,
    get_registry_address,
)
from lido.contracts.abi_loader import load_lido_abi, load_operators_abi


def get_lido_contract(address: t.Optional[str] = None, path: t.Optional[str] = None):
    """Load Lido web3 contract object"""
    return w3.eth.contract(address=get_lido_address(address=address), abi=load_lido_abi(path=path))


def get_nos_contract(address: t.Optional[str] = None, path: t.Optional[str] = None):
    """Load Node Operator web3 contract object"""
    return w3.eth.contract(
        address=get_registry_address(address=address), abi=load_operators_abi(path=path)
    )



from typing import Dict, NamedTuple


DEPOSIT_CLI_VERSION = "1.1.0"


class BaseChainSetting(NamedTuple):
    ETH2_NETWORK_NAME: str
    GENESIS_FORK_VERSION: bytes


MAINNET = "mainnet"
WITTI = "witti"
ALTONA = "altona"
MEDALLA = "medalla"
SPADINA = "spadina"
ZINKEN = "zinken"
PYRMONT = "pyrmont"


# Eth2 Mainnet setting
MainnetSetting = BaseChainSetting(
    ETH2_NETWORK_NAME=MAINNET, GENESIS_FORK_VERSION=bytes.fromhex("00000000")
)
# Eth2 spec v0.11.3 testnet
WittiSetting = BaseChainSetting(
    ETH2_NETWORK_NAME=WITTI, GENESIS_FORK_VERSION=bytes.fromhex("00000113")
)
# Eth2 spec v0.12.1 testnet
AltonaSetting = BaseChainSetting(
    ETH2_NETWORK_NAME=ALTONA, GENESIS_FORK_VERSION=bytes.fromhex("00000121")
)
# Eth2 "official" public testnet (spec v0.12.2)
MedallaSetting = BaseChainSetting(
    ETH2_NETWORK_NAME=MEDALLA, GENESIS_FORK_VERSION=bytes.fromhex("00000001")
)
# Eth2 "dress rehearsal" testnet (spec v0.12.3)
SpadinaSetting = BaseChainSetting(
    ETH2_NETWORK_NAME=SPADINA, GENESIS_FORK_VERSION=bytes.fromhex("00000002")
)
# Eth2 "dress rehearsal" testnet (spec v0.12.3)
ZinkenSetting = BaseChainSetting(
    ETH2_NETWORK_NAME=ZINKEN, GENESIS_FORK_VERSION=bytes.fromhex("00000003")
)
# Eth2 pre-launch testnet (spec v1.0.0)
PyrmontSetting = BaseChainSetting(
    ETH2_NETWORK_NAME=PYRMONT, GENESIS_FORK_VERSION=bytes.fromhex("00002009")
)


ALL_CHAINS: Dict[str, BaseChainSetting] = {
    MAINNET: MainnetSetting,
    WITTI: WittiSetting,
    ALTONA: AltonaSetting,
    MEDALLA: MedallaSetting,
    SPADINA: SpadinaSetting,
    ZINKEN: ZinkenSetting,
    PYRMONT: PyrmontSetting,
}


def get_chain_setting(chain_name: str = MAINNET) -> BaseChainSetting:
    return ALL_CHAINS[chain_name]



# SPDX-FileCopyrightText: 2020 Lido <info@lido.fi>

# SPDX-License-Identifier: GPL-3.0

import base64
import binascii
import datetime
import logging
import math
from datetime import timezone

import requests
from requests.compat import urljoin

logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)8s %(asctime)s <daemon> %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


def get_beacon(provider, slots_per_epoch):
    version = requests.get(urljoin(provider, "eth/v1/node/version")).text
    if "Lighthouse" in version:
        return Lighthouse(provider, slots_per_epoch)
    version = requests.get(urljoin(provider, "eth/v1alpha1/node/version")).text
    if "Prysm" in version:
        return Prysm(provider, slots_per_epoch)
    raise ValueError("Unknown beacon")


class Lighthouse:
    api_version = "eth/v1/node/version"
    api_genesis = "eth/v1/beacon/genesis"
    api_beacon_head_finality_checkpoints = "eth/v1/beacon/states/head/finality_checkpoints"
    api_beacon_head_finalized = "eth/v1/beacon/headers/finalized"
    api_beacon_head_actual = "eth/v1/beacon/headers/head"
    api_get_balances = "eth/v1/beacon/states/{}/validators"
    api_get_slot = "eth/v1/beacon/states/{}/root"

    def __init__(self, url, slots_per_epoch):
        self.url = url
        self.slots_per_epoch = slots_per_epoch
        self.version = requests.get(urljoin(url, self.api_version)).json()

    def get_finalized_epoch(self):
        return int(
            requests.get(urljoin(self.url, self.api_beacon_head_finality_checkpoints)).json()[
                "data"
            ]["finalized"]["epoch"]
        )

    def get_genesis(self):
        return int(requests.get(urljoin(self.url, self.api_genesis)).json()["data"]["genesis_time"])

    def get_actual_slot(self):
        actual_slots = {}
        response = requests.get(urljoin(self.url, self.api_beacon_head_actual)).json()
        actual_slots["actual_slot"] = int(response["data"]["header"]["message"]["slot"])
        response = requests.get(urljoin(self.url, self.api_beacon_head_finalized)).json()
        actual_slots["finalized_slot"] = int(response["data"]["header"]["message"]["slot"])
        return actual_slots

    def _convert_key_list_to_str_arr(self, key_list):
        pubkeys = []
        for key in key_list:
            pubkeys.append("0x" + binascii.hexlify(key).decode())

        return pubkeys

    def get_balances(self, slot, key_list):
        pubkeys = self._convert_key_list_to_str_arr(key_list)

        logging.info("Fetching validators from Beacon node...")
        response_json = requests.get(urljoin(self.url, self.api_get_balances.format(slot))).json()
        balance_list = []
        found_on_beacon_pubkeys = []
        logging.info(f"Validator balances on beacon for slot: {slot}")
        active_validators_balance = 0
        for validator in response_json["data"]:
            pubkey = validator["validator"]["pubkey"]
            # Log all validators along with balance
            if pubkey in pubkeys:
                validator_balance = int(validator["balance"])

                if validator["status"] == "active":
                    active_validators_balance += validator_balance

                balance_list.append(validator_balance)
                found_on_beacon_pubkeys.append(validator["validator"]["pubkey"])
                logging.info(f"Pubkey: {pubkey[:12]} Balance: {validator_balance} Gwei")
            elif validator["status"] == "UNKNOWN":
                logging.warning(f"Pubkey {pubkey[:12]} status UNKNOWN")
        balance = sum(balance_list)

        # Convert Gwei to wei
        balance *= 10 ** 9
        active_validators_balance *= 10 ** 9

        validators = len(found_on_beacon_pubkeys)
        return balance, validators, active_validators_balance


class Prysm:
    api_version = "eth/v1alpha1/node/version"
    api_genesis = "eth/v1alpha1/node/genesis"
    api_beacon_head = "eth/v1alpha1/beacon/chainhead"
    api_get_balances = "eth/v1alpha1/validators/balances"

    def __init__(self, url, slots_per_epoch):
        self.url = url
        self.slots_per_epoch = slots_per_epoch
        self.version = requests.get(urljoin(url, self.api_version)).json()

    def get_finalized_epoch(self):
        finalized_epoch = int(
            requests.get(urljoin(self.url, self.api_beacon_head)).json()["finalizedEpoch"]
        )
        return finalized_epoch

    def get_genesis(self):
        genesis_time = requests.get(urljoin(self.url, self.api_genesis)).json()["genesisTime"]
        genesis_time = datetime.datetime.strptime(genesis_time, "%Y-%m-%dT%H:%M:%SZ").replace(
            tzinfo=timezone.utc
        )
        genesis_time = int(genesis_time.timestamp())
        return genesis_time

    def get_actual_slot(self):
        actual_slots = {}
        response = requests.get(urljoin(self.url, self.api_beacon_head)).json()
        actual_slots["actual_slot"] = int(response["headSlot"])
        actual_slots["finalized_slot"] = int(response["finalizedSlot"])
        return actual_slots

    def get_balances(self, slot, key_list):
        params = {}
        pubkeys = []
        found_on_beacon_pubkeys = []
        balance_list = []
        key_dict = {}
        for key in key_list:
            base64_key = base64.b64encode(key).decode()
            hex_key = "0x" + binascii.hexlify(key).decode()
            key_dict[base64_key] = hex_key[:12]
            pubkeys.append(base64_key)

        epoch = math.ceil(slot / self.slots_per_epoch)  # Round up in case of missing slots

        active_validators_balance = 0

        for pk in pubkeys:
            params["publicKeys"] = pk
            params["epoch"] = epoch
            response = requests.get(urljoin(self.url, self.api_get_balances), params=params)
            if "error" in response.json():
                logging.error(f"Pubkey {key_dict[pk]} return error")
                continue
            validator = response.json()["balances"][0]

            if validator["publicKey"] in pubkeys:
                found_on_beacon_pubkeys.append(validator["publicKey"])
                balance = int(validator["balance"])
                if validator["status"] == "ACTIVE":
                    active_validators_balance += balance

                balance_list.append(balance)
                logging.info(f"Pubkey: {key_dict[pk]} Balance: {balance} Gwei")
            elif validator["status"] == "UNKNOWN":
                logging.warning(f"Pubkey {key_dict[pk]} status UNKNOWN")

        balances = sum(balance_list)
        # Convert Gwei to wei
        balances *= 10 ** 9
        active_validators_balance *= 10 ** 9
        total_validators_on_beacon = len(found_on_beacon_pubkeys)

        return balances, total_validators_on_beacon, active_validators_balance



import typing as t
import logging

from lido.multicall import Call, Multicall
from lido.constants.contract_addresses import get_registry_address
from lido.contracts.w3_contracts import get_nos_contract

logger = logging.getLogger(__name__)

multicall_default_batch = 300


def get_operators_keys(
    operators: t.List[t.Dict],
    max_multicall: t.Optional[int] = None,
    registry_address: t.Optional[str] = None,
    registry_abi_path: t.Optional[str] = None,
) -> t.List[t.Dict]:
    """Get and add signing keys to node operators

    Example output:
    [{
        'id': 0,
        'active': True,
        'name': 'Staking Facilities',
        'rewardAddress': '0xdd4bc51496dc93a0c47008e820e0d80745476f22',
        'stakingLimit': 1000,
        'stoppedValidators': 0,
        'totalSigningKeys': 1000,
        'usedSigningKeys': 560,
        'keys': [{
            'index': 0,
            'key': b '\x81\xb4\xaea\xa8\x989i\x03\x89\x7f\x94\xbe\xa0\xe0b\xc3\xa6\x92^\xe9=0\xf4\xd4\xae\xe9;S;IU\x1a\xc37\xdax\xff*\xb0\xcf\xbb\n\xdb8\x0c\xad\x94',
            'depositSignature': b '\x96\xa8\x8f\x8e\x88>\x9f\xf6\xf3\x97Q\xa2\xcb\xfc\xa3\x94\x9fO\xa9j\xe9\x84D\xd0\x05\xb6\xea\x9f\xaa\xc0\xc3KR\xd5\x95\xf9B\x8d\x90\x1d\xdd\x815$\x83}\x86d\x01\xedL\xed=\x84\xe7\xe88\xa2e\x06\xae.\xf3\xbf\x0b\xf1\xb8\xd3\x8b+\xd7\xbd\xb6\xc1<_F\xb8H\xd0-\xdc\x11\x08d\x9e\x96\x07\xcfM/\xce\xcd\xd8\x07\xbb',
            'used': True
        }, ...]
    }, ...]
    """

    # Default multicall batch size
    if max_multicall is None:
        max_multicall = multicall_default_batch

    address = registry_address or get_registry_address()

    for op_i, op in enumerate(operators):
        total_keys = op["totalSigningKeys"]

        keys = []

        calls_list = []
        for i in range(total_keys):
            call = Call(
                address,
                [
                    "getSigningKey(uint256,uint256)(bytes,bytes,bool)",
                    op_i,
                    i,
                ],
                [[i, None]],
            )
            calls_list.append(call)
            if len(calls_list) >= max_multicall:
                logger.debug(f"{len(calls_list)=}")
                multi_call = Multicall(calls_list)()
                calls_list = []
                keys.extend(multi_call.values())

        if calls_list:
            logger.debug(f"{len(calls_list)=}")
            multi_call = Multicall(calls_list)()
            keys.extend(multi_call.values())

        function_abi = next(
            x
            for x in get_nos_contract(address=registry_address, path=registry_abi_path).abi
            if x["name"] == "getSigningKey"
        )
        signing_keys_keys = ["index"] + [x["name"] for x in function_abi["outputs"]]
        signing_keys_list = [
            dict(
                zip(
                    signing_keys_keys,
                    [i] + list(item),
                )
            )
            for i, item in enumerate(keys)
        ]

        operators[op_i]["keys"] = signing_keys_list

    return operators



import typing as t


def spot_duplicates(
    operators: t.List[t.Dict], key: t.Dict, original_op: t.Optional[t.Dict] = None
) -> t.List:
    """Compare every available key with other keys to spot duplicates"""

    duplicates_found = []
    for operator in operators:
        for second_key in operator["keys"]:
            if key["key"] == second_key["key"]:
                if not original_op:
                    duplicates_found.append({"op": operator, "key": second_key})
                elif (original_op["id"], key["index"]) != (
                    operator["id"],
                    second_key["index"],
                ):
                    duplicates_found.append({"op": operator, "key": second_key})
    return duplicates_found


def find_duplicates(operators: t.List[t.Dict]) -> t.List[t.Dict]:
    """Loop through all keys and
    add information about duplicates to "duplicate" and "duplicates" fields of every key"""

    for op_i, op in enumerate(operators):
        for key_i, key in enumerate(op["keys"]):
            operators[op_i]["keys"][key_i]["duplicates"] = []

            duplicates = spot_duplicates(operators, key, op)
            for duplicate_i, duplicate in enumerate(duplicates):
                operators[op_i]["keys"][key_i]["duplicates"].append(
                    dict(
                        op_id=duplicate["op"]["id"],
                        op_name=duplicate["op"]["name"],
                        index=duplicate["key"]["index"],
                        approved=bool(duplicate["op"]["stakingLimit"]),
                        used=duplicate["key"]["used"],
                    )
                )

            if duplicates:
                operators[op_i]["keys"][key_i]["duplicate"] = True
            else:
                operators[op_i]["keys"][key_i]["duplicate"] = False

    return operators



import typing as t

import os

from web3.auto import w3
from lido.constants.chains import chains

LIDO_ADDRESSES = {
    "mainnet": "0xae7ab96520DE3A18E5e111B5EaAb095312D7fE84",
    "goerli": "0xA5d26F68130c989ef3e063c9bdE33BC50a86629D",
}

NODE_OPS_ADDRESSES = {
    "mainnet": "0x55032650b14df07b85bF18A3a3eC8E0Af2e028d5",
    "goerli": "0xB1e7Fb9E9A71063ab552dDEE87Ea8C6eEc7F5c7A",
}


def get_lido_address(address: t.Optional[str] = None) -> str:
    """Return an appropriate Lido address for current network"""

    if address:
        return address

    env = os.getenv("LIDO_ADDRESS")

    if env is not None:
        return env

    chain = w3.eth.chainId
    return LIDO_ADDRESSES[chains[chain]]


def get_registry_address(address: t.Optional[str] = None) -> str:
    """Return an appropriate Node Operator (registry) address for current network"""

    if address:
        return address

    env = os.getenv("REGISTRY_ADDRESS")

    if env is not None:
        return env

    chain = w3.eth.chainId
    return NODE_OPS_ADDRESSES[chains[chain]]



import typing as t

from lido import (
    get_operators_data,
    get_operators_keys,
    validate_keys_multi,
    find_duplicates,
)


def fetch_and_validate(
    max_multicall: t.Optional[int] = None,
    lido_address: t.Optional[str] = None,
    lido_abi_path: t.Optional[str] = None,
    registry_address: t.Optional[str] = None,
    registry_abi_path: t.Optional[str] = None,
) -> t.List[t.Dict]:
    """fetches all operator keys, and run keys validations and duplication checks

    check fields:
     - operators[op_i]["keys"][key_i]["duplicate"]
     - operators[op_i]["keys"][key_i]["duplicates"]
     - operators[op_i]["keys"][key_i]["valid_signature"]
    """

    operators_data = get_operators_data(
        registry_address=registry_address, registry_abi_path=registry_abi_path
    )

    data_with_keys = get_operators_keys(
        operators=operators_data,
        max_multicall=max_multicall,
        registry_address=registry_address,
        registry_abi_path=registry_abi_path,
    )

    data_validated_keys = validate_keys_multi(
        operators=data_with_keys, lido_address=lido_address, lido_abi_path=lido_abi_path
    )

    data_found_duplicates = find_duplicates(operators=data_validated_keys)

    return data_found_duplicates



import typing as t
import logging

from lido.multicall import Call, Multicall
from lido.constants.contract_addresses import get_registry_address
from lido.contracts.w3_contracts import get_nos_contract

logger = logging.getLogger(__name__)


def get_operators_data(
    registry_address: t.Optional[str] = None,
    registry_abi_path: t.Optional[str] = None,
) -> t.List[t.Dict]:
    """Fetch information for each node operator"""
    address = registry_address or get_registry_address()
    operators_n = Call(address, "getNodeOperatorsCount()(uint256)")()
    logger.debug(f'{operators_n=}')
    if operators_n == 0:
        logger.warning(f'no operators')  # fixme assert if not test env
        return []
    assert operators_n < 1_000_000, 'too big operators_n'

    calls = Multicall(
        [
            Call(
                address,
                [
                    "getNodeOperator(uint256,bool)(bool,string,address,uint64,uint64,uint64,uint64)",
                    i,
                    True,
                ],
                [[i, None]],
            )
            for i in range(operators_n)
        ]
    )()

    calls_list = list(calls.values())

    # Adding index as first data of operator
    calls_with_indeces = [[i] + list(item) for i, item in enumerate(calls_list)]

    # Getting function data from contract ABI
    function_abi = next(
        x
        for x in get_nos_contract(address=registry_address, path=registry_abi_path).abi
        if x["name"] == "getNodeOperator"
    )

    # Adding "id" and the rest of output name keys
    op_keys = ["id"] + [x["name"] for x in function_abi["outputs"]]
    operators = [dict(zip(op_keys, op)) for op in calls_with_indeces]

    return operators



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lido',
 'lido.constants',
 'lido.contracts',
 'lido.eth2deposit',
 'lido.eth2deposit.utils',
 'lido.multicall',
 'lido.utils']

package_data = \
{'': ['*'], 'lido.contracts': ['abi/goerli/*', 'abi/mainnet/*']}

install_requires = \
['py-ecc>=5.1.0,<6.0.0',
 'requests>=2.25.1,<3.0.0',
 'ssz>=0.2.4,<0.3.0',
 'web3>=5.15.0,<6.0.0']

setup_kwargs = {
    'name': 'lido',
    'version': '0.2.2',
    'description': 'Network helpers for Lido',
    'long_description': '# Lido\n\nThis library consolidates various functions to efficiently load network data for Lido, validate node operator keys and find key duplicates.\n\n## Installation\n\nThis library is available on PyPi:\n\n`pip install lido`\n\n## Main Features\n\n### Multicall Function Calls\n\nInstead of making network requests one-by-one, this library combines many requests into one RPC call. It uses [banteg/multicall.py](https://github.com/banteg/multicall.py), a Python wrapper for [makerdao/multicall](https://github.com/makerdao/multicall).\n\n### Multiprocess Signature Validations\n\nWhen using `validate_keys_multi()`, this library spreads processing of key signature validations to all system cores.\n\n### Automatic Testnet / Mainnet Switching\n\nDepending on the supplied WEB3_PROVIDER_URI, a correct network will be used. Even an appropriate ABI will be loaded for the chain automatically.\n\n## Helpers Provided\n\n- get_operators_data() -> operator_data - load node operator data\n\n- get_operators_keys(operator_data) -> operator_data - fetches and adds keys to operator_data\n- validate_keys_mono(operator_data) -> operator_data - validates keys in single process and adds validation results to operator_data\n- validate_keys_multi(operator_data) -> operator_data - validates keys in multiple processes and adds validation results to operator_data, requires a main function (see example)\n- validate_key([[key,depositSignature]]) -> Boolean - low-level validation function\n- find_duplicates(operator_data) -> operator_data - finds duplicate keys and adds results to operator_data\n\n- fetch_and_validate() -> operator_data - combines fetching operator data and running all validations on it - useful when you would be running all validations on data anyway\n\n- get_stats() -> stats - fetches various constants from Lido contract, but you can even pass a list of functions to fetch eg get_stats([isStopped])\n\nYou can mix and match these functions, but make sure to use get_operators_data() first.\n\n## Notes\n\n1. Signature validation will be skipped if its results are already present in operator_data. This way you can safely load validation results from cache and add `["valid_signature"] = Boolean` to already checked keys.\n\n## How to Use\n\nUse a RPC provider url as an environment variable and run your script:\n\n`WEB3_PROVIDER_URI=https://eth-mainnet.provider.xx example.py`\n\nSee `example.py` for a complete example, make sure to use a main function and a script entry point check when using validate_keys_multi() or fetch_and_validate().\n\n## Options\n\nIf you are testing a new deployment of Lido, these environment variables can override addresses and ABIs:\n\n- LIDO_ADDRESS\n- REGISTRY_ADDRESS\n- LIDO_ABI (the file-path to the contract\'s ABI)\n- REGISTRY_ABI (the file-path to the contract\'s ABI)\n\n`WEB3_PROVIDER_URI=https://eth-mainnet.provider.xx LIDO_ADDRESS=XXX REGISTRY_ADDRESS=XXX LIDO_ABI=xxx.json REGISTRY_ABI=xxx.json example.py`\n',
    'author': 'Lido',
    'author_email': 'info@lido.fi',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://lido.fi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>3.7.1,<4',
}


setup(**setup_kwargs)



import typing as t

import os
import json
from lido.constants.chains import get_chain_name

network = get_chain_name()
script_dir = os.path.dirname(__file__)


def load_lido_abi(path: t.Optional[str] = None) -> t.List[t.Dict]:
    """Load an appropriate ABI file for Lido"""

    if path:
        return json.load(open(path))

    env = os.getenv("lido_abi")

    if env is not None:
        return json.load(open(env))

    if network == "mainnet":
        return json.load(open(os.path.join(script_dir, "abi/mainnet/Lido.json")))
    elif network == "goerli":
        return json.load(open(os.path.join(script_dir, "abi/goerli/Lido.json")))
    else:
        raise Exception("Unable to load Lido ABI for network")


def load_operators_abi(path: t.Optional[str] = None) -> t.List[t.Dict]:
    """Load an appropriate ABI file for Node Operators"""

    if path:
        return json.load(open(path))

    env = os.getenv("REGISTRY_ABI")

    if env is not None:
        return json.load(open(env))

    if network == "mainnet":
        return json.load(open(os.path.join(script_dir, "abi/mainnet/NodeOperatorsRegistry.json")))
    elif network == "goerli":
        return json.load(open(os.path.join(script_dir, "abi/goerli/NodeOperatorsRegistry.json")))
    else:
        raise Exception("Unable to load NodeOperatorsRegistry ABI for network")
